/**************************************************************************************************************************
// This script was authored by Lorne Brooks - Email: lbtools2k@gmail.com
//
//
// REVISION HISTORY:
// Version 1.01 12/08/23;
//						- Hotfix: prevented crash due to struct reference error.
// Version 1.0 11/08/23;
//						- First public release.
//
//
// This script is not copyrighted, please feel free to modify or distribute.
//
**************************************************************************************************************************/


global g_LBTools_quickSel

struct s_LBTools_quickSelection
(

/*==================================================================================
// Structure variables
====================================================================================*/

quickSelection_Rollout,
about_Rollout,
QS_pos,
qsINIfile = (getDir #plugcfg) + @"\LBTools_QuickSelection.ini",
textFile = @"C:\TEMP\LBTools_QuickSelection.txt",
imageGuide = (getDir #usericons) + @"\LBTools_Guide_Image.bmp",
imageNotepad = (getDir #usericons) + @"\LBTools_Notepad_Image.bmp",
pathsButtonState = false,
a_storedObjs = #(), --// an array used to store nodes in memory for later retrieval
a_setItems = #(), --// an array of selection set names stored in the INI file
a_pathItems = #("*** EMPTY ***"), --// an array of strings for finding objects using wildcards stored in the text file
flag_stored = false,
flag_hidden = false,
flag_frozen = false,
borderActive = bitMap 155 32 color:[0,250,180],
borderInactive = bitMap 155 32 color:gray,
pathsBorderState = borderInactive,
	
/*==================================================================================
// Structure functions
/*==================================================================================

f_initializeIniFile
f_loadSelectionSets
f_loadPathItemsFromFile
f_createPathsTextFile
f_selectObjects
f_selectSetObjects
f_selectPathObjects
f_quickSelection
	
====================================================================================*/

fn f_initializeIniFile =
(
	if NOT (doesFileExist g_LBTools_quickSel.qsINIfile) then
	(
		g_LBTools_quickSel.QS_pos = mouse.screenpos + [-75,-15]
		setIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "Position" (g_LBTools_quickSel.QS_pos as string)
		setIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "IncludeHidden" (g_LBTools_quickSel.flag_hidden as string)
		setIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "IncludeFrozen" (g_LBTools_quickSel.flag_frozen as string)
	)
	else
	(
		g_LBTools_quickSel.QS_pos = execute (getIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "Position")
		g_LBTools_quickSel.flag_hidden = execute (getIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "IncludeHidden")
		g_LBTools_quickSel.flag_frozen = execute (getIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "IncludeFrozen")
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_loadSelectionSets =
(
	if (hasIniSetting g_LBTools_quickSel.qsINIfile "SelectionSets") then
	(
		g_LBTools_quickSel.a_setItems = quickSelection_Rollout.ddl_sets.items = getIniSetting g_LBTools_quickSel.qsINIfile "SelectionSets"
	)
	else
	(
		g_LBTools_quickSel.a_setItems = quickSelection_Rollout.ddl_sets.items = #()
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_loadPathItemsFromFile =
(
	g_LBTools_quickSel.a_pathItems = #()
	local textFileRead = openFile g_LBTools_quickSel.textFile mode:"r"
	while NOT eof textFileRead do
	(
		local str_line = readLine textFileRead
		if str_line != "" then --// Ignore blank lines
		(
			append g_LBTools_quickSel.a_pathItems str_line
		)
	)
	close textFileRead
	if g_LBTools_quickSel.a_pathItems.count == 0 then
	(
		g_LBTools_quickSel.a_pathItems = quickSelection_Rollout.ddl_paths.items = #("*** EMPTY ***")
		quickSelection_Rollout.btn_grabPathObjs.enabled = g_LBTools_quickSel.pathsButtonState = false
		quickSelection_Rollout.borderImagePaths.bitMap = g_LBTools_quickSel.pathsBorderState = g_LBTools_quickSel.borderInactive
	)
	else
	(
		quickSelection_Rollout.ddl_paths.items = g_LBTools_quickSel.a_pathItems
		quickSelection_Rollout.btn_grabPathObjs.enabled = g_LBTools_quickSel.pathsButtonState = true
		quickSelection_Rollout.borderImagePaths.bitMap = g_LBTools_quickSel.pathsBorderState = g_LBTools_quickSel.borderActive
		if g_LBTools_quickSel.a_pathItems.count == 1 AND g_LBTools_quickSel.a_pathItems[1] == "*** EMPTY ***" then
		(
			quickSelection_Rollout.btn_grabPathObjs.enabled = g_LBTools_quickSel.pathsButtonState = false
			quickSelection_Rollout.borderImagePaths.bitMap = g_LBTools_quickSel.pathsBorderState = g_LBTools_quickSel.borderInactive
		)
	) 	
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_createPathsTextFile =
(
	if NOT doesFileExist g_LBTools_quickSel.textFile then
	(
		makeDir @"C:\Temp"
		local edit_txtFile = openFile g_LBTools_quickSel.textFile mode:"w"
		format "*** EMPTY ***\n" to:edit_txtFile
		close edit_txtFile
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_selectObjects a_objs =
(
	local a_objsToSelect = #()
	
	for obj in a_objs do
	(
		if isValidNode obj then
		(
			local allowSelection = false
			
			if NOT obj.isHidden AND NOT obj.isFrozen then
			(
				allowSelection = true
			)
			else if obj.isHidden AND g_LBTools_quickSel.flag_hidden then
			(
				allowSelection = true
			)
			else if NOT obj.isHidden AND obj.isFrozen AND g_LBTools_quickSel.flag_frozen then
			(
				allowSelection = true
			)

			if allowSelection then append a_objsToSelect obj
		)
	)
	
	if a_objsToSelect.count > 0 then select a_objsToSelect
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_selectSetObjects setName =
(
	local a_setObjects = #()
	local a_objNames = execute(getIniSetting g_LBTools_quickSel.qsINIfile "SelectionSets" setName)
	if a_objNames.count > 0 then
	(
		for objName in a_objNames do
		(
			local obj = getNodeByName objName exact:true
			if isValidNode obj then
			(
				append a_setObjects obj
			)
		)
		g_LBTools_quickSel.f_selectObjects a_setObjects
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_selectPathObjects id =
(
	local str_item = g_LBTools_quickSel.a_pathItems[id]
	if str_item != undefined AND str_item != "*** EMPTY ***" then
	(
		local ss_path = stringstream ""
		format "$%" str_item to:ss_path
		local str_path = ss_path as string
		--// Note: The onus is on the user to format their pathname literals correctly
		try(
			g_LBTools_quickSel.f_selectObjects (execute str_path)
		)catch(
			local ss_msg = stringstream ""
			format "Error: Pathname literal not valid for...\n\n%\n\nCheck that names with spaces are between single quotes." str_path to:ss_msg
			messagebox (ss_msg as string) title:"Error"
		)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_quickSelection =
(
	rollout quickSelection_Rollout "Quick Selection"
	(
		group "Settings"
		(
			checkbox chkbx_hidden "Select hidden objects" checked:(g_LBTools_quickSel.flag_hidden) tooltip:"Allow hidden objects to be included in the selection (frozen or not)"
			checkbox chkbx_frozen "Select visible frozen objects" checked:(g_LBTools_quickSel.flag_frozen) tooltip:"Allow visible frozen objects to be included in the selection"
			button btn_hideSelection "Hide" align:#center width:70 height:20 across:2 tooltip:"Set all selected objects to hidden"
			button btn_unhideSelection "Unhide" align:#center width:70 height:20 tooltip:"Set all selected objects to unhidden"
			button btn_freezeSelection "Freeze" align:#center width:70 height:20 across:2 tooltip:"Set all selected objects to frozen"
			button btn_unfreezeSelection "Unfreeze" align:#center width:70 height:20 tooltip:"Set all selected objects to unfrozen"
		)
		group "Category Selection"
		(
			button btn_geometry "Geometry" width:70 height:20 align:#left offset:[2,0] across:2
			button btn_shapes "Shapes" width:70 height:20 align:#right offset:[-2,0]
			button btn_lights "Lights" width:70 height:20 align:#left offset:[2,0] across:2
			button btn_cameras "Cameras" width:70 height:20 align:#right offset:[-2,0]
			button btn_helpers "Helpers" width:70 height:20 align:#left offset:[2,0] across:2
			button btn_bones "Bones" width:70 height:20 align:#right offset:[-2,0]
		)
		group "Stored Selection"
		(
			button btn_storeSelection "Store" align:#center width:152 height:20 tooltip:"Store the selection of objects so they can be selected again (from memory)"
			bitmap borderImageStored bitMap:(g_LBTools_quickSel.borderInactive)
			button btn_grabStoredObjs "Grab Stored Objects" enabled:false align:#center width:145 height:25 offset:[0,-40] tooltip:"Selects the objects that were stored"
			label llb_gap1 ""
		)
		group "Selection Sets"
		(
			edittext edtxt_setName "Name: " align:#left fieldwidth:115 tooltip:"Enter a name for a new selection set to store"
			button btn_addSet "Add" ailgn:#left width:70 height:20 across:2 offset:[-2,0] tooltip:"Store the selected objects as a new selection set using the name provided (saved to file)"
			button btn_removeSet "Remove" align:#right width:70 height:20 offset:[-4,0] tooltip:"Removes the picked selection set from the list"
			dropdownlist ddl_sets "" items:(g_LBTools_quickSel.a_setItems) align:#center width:150 height:20 tooltip:"Choose a selection set to select objects from"
			bitmap borderImageSets bitMap:(g_LBTools_quickSel.borderInactive)
			button btn_grabSetObjs "Grab Set Objects" enabled:false ailgn:#center width:145 height:25 offset:[0,-40] tooltip:"Select the objects for the picked selection set"
			label llb_gap2 ""
		)
		group "Path Selection"
		(
			button btn_refresh "Refresh List" align:#left width:100 height:20 offset:[25,0] across:2 tooltip:"Refresh the list with text file contents"
			button btn_openFile "" images:#(g_LBTools_quickSel.imageNotepad, undefined, 1, 1, 1, 1, 1, false) align:#right width:18 height:20 offset:[-6,0] tooltip:"Open a text file in an editor to add path strings to the list"
			label lbl_path "$" align:#left offset:[2,4] across:2
			dropdownlist ddl_paths "" items:(g_LBTools_quickSel.a_pathItems) align:#right width:147 height:20 offset:[0,0] tooltip:"Choose a path definition to select objects from"
			bitmap borderImagePaths bitMap:(g_LBTools_quickSel.pathsBorderState)
			button btn_grabPathObjs "Grab Path Objects" enabled:(g_LBTools_quickSel.pathsButtonState) align:#center width:145 height:25 offset:[0,-40] tooltip:"Select the objects for the current paths list item"
			label llb_gap3 ""
		)
		button btn_about "About" align:#center width:40 height:20
		
		on quickSelection_Rollout open do
		(
			if g_LBTools_quickSel.flag_stored then
			(
				btn_grabStoredObjs.enabled = true
			)
			if g_LBTools_quickSel.a_setItems.count > 0 then
			(
				btn_grabSetObjs.enabled = true
				borderImageSets.bitmap = g_LBTools_quickSel.borderActive
			)
		)
			
		on quickSelection_Rollout moved pos do
		(
			g_LBTools_quickSel.QS_Pos = pos
			setIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "Position" (g_LBTools_quickSel.QS_pos as string)
		)
		
		on quickSelection_Rollout close do
		(
			try(destroyDialog g_LBTools_quickSel.about_Rollout)catch()
		)
		
		on btn_about pressed do
		(
			try(destroyDialog g_LBTools_quickSel.about_Rollout)catch()
			createDialog g_LBTools_quickSel.about_Rollout
		)
		
		on btn_geometry pressed do
		(
			local a_geometry = for g in geometry where classof g != BoneGeometry collect g
			g_LBTools_quickSel.f_selectObjects a_geometry
		)
		on btn_shapes pressed do g_LBTools_quickSel.f_selectObjects shapes
		on btn_lights pressed do g_LBTools_quickSel.f_selectObjects lights
		on btn_cameras pressed do g_LBTools_quickSel.f_selectObjects cameras
		on btn_helpers pressed do g_LBTools_quickSel.f_selectObjects helpers
		on btn_bones pressed do
		(
			local a_bones = for o in objects where classof o == BoneGeometry collect o
			g_LBTools_quickSel.f_selectObjects a_bones
		)

		on chkbx_hidden changed state do
		(
			g_LBTools_quickSel.flag_hidden = state
			setIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "IncludeHidden" (g_LBTools_quickSel.flag_hidden as string)
		)
		
		on chkbx_frozen changed state do
		(
			g_LBTools_quickSel.flag_frozen = state
			setIniSetting g_LBTools_quickSel.qsINIfile "QuickSelection" "IncludeFrozen" (g_LBTools_quickSel.flag_frozen as string)
		)
		
		on btn_hideSelection pressed do
		(
			undo "HideSelection" on hide selection
		)
		
		on btn_unhideSelection pressed do
		(
			undo "UnhideSelection" on unhide selection
		)
		
		on btn_freezeSelection pressed do
		(
			undo "FreezeSelection" on freeze selection
		)
		
		on btn_unfreezeSelection pressed do
		(
			undo "UnfreezeSelection" on unfreeze selection
		)
		
		--// STORED SELECTION (memory)
		on btn_storeSelection pressed do
		(
			if selection.count > 0 then
			(
				g_LBTools_quickSel.a_storedObjs = for s in selection collect s
				g_LBTools_quickSel.flag_stored = true
				btn_grabStoredObjs.enabled = true
				borderImageStored.bitmap = g_LBTools_quickSel.borderActive
			)
			else
			(
				messageBox "Error: Select some objects to store first." title:"Error"
			)
		)
				
		on btn_grabStoredObjs pressed do
		(
			if g_LBTools_quickSel.a_storedObjs.count > 0 then
			(
				local a_foundObjs = #()
				for o in g_LBTools_quickSel.a_storedObjs do
				(
					if isValidNode o then append a_foundObjs o
				)
				if a_foundObjs.count > 0 then
				(
					g_LBTools_quickSel.f_selectObjects a_foundObjs
				)
				else
				(
					messageBox "Error: No stored objects were found. They may have been deleted." title:"Error"
					g_LBTools_quickSel.flag_stored = false
					g_LBTools_quickSel.a_storedObjs = #()
					btn_grabStoredObjs.enabled = false
					borderImageStored.bitmap = g_LBTools_quickSel.borderInactive
				)	
			)
			else
			(
				messageBox "Error: Nothing is stored." title:"Error"
			)
		)

		--// STORED SELECTION SETS (file)
		on btn_addSet pressed do
		(
			if edtxt_setName.text != "" then
			(
				local a_filtered = filterstring edtxt_setName.text " "
				if a_filtered.count > 0 then
				(					
					if selection.count > 0 then
					(
						local a_objNames = for obj in selection collect obj.name
						setIniSetting g_LBTools_quickSel.qsINIfile "SelectionSets" edtxt_setName.text (a_objNames as string)
						g_LBTools_quickSel.f_loadSelectionSets()
						ddl_sets.selection = ddl_sets.items.count
						btn_grabSetObjs.enabled = true
						borderImageSets.bitmap = g_LBTools_quickSel.borderActive
					)
					else
					(
						messagebox "Error: Must have one or more objects selected." title:"Error"
						setFocus edtxt_setName
					)
				)
				else
				(
					messagebox "Error: Must enter a selection set name (and no white space only)." title:"Error"
					edtxt_setName.text = ""
					setFocus edtxt_setName
				)
			)
			else
			(
				messagebox "Error: Must enter a selection set name." title:"Error"
				setFocus edtxt_setName
			)
		)
		
		on btn_removeSet pressed do
		(
			if ddl_sets.items.count > 0 AND (querybox "Are you sure you want to remove the selection set?" title:"Quick Selection") then
			(
				local sets = getIniSetting g_LBTools_quickSel.qsINIfile "SelectionSets"
				delIniSetting g_LBTools_quickSel.qsINIfile "SelectionSets" sets[ddl_sets.selection]
				g_LBTools_quickSel.f_loadSelectionSets()
				if ddl_sets.items.count > 0 then
				(
					ddl_sets.selection = 1
				)
				else
				(
					btn_grabSetObjs.enabled = false
					borderImageSets.bitmap = g_LBTools_quickSel.borderInactive
				)
			)
		)
		
		on ddl_sets selected val do
		(
			g_LBTools_quickSel.f_selectSetObjects ddl_sets.items[val]
		)
		
		on btn_grabSetObjs pressed do
		(
			g_LBTools_quickSel.f_selectSetObjects ddl_sets.items[ddl_sets.selection]
		)
		
		--// PATH SELECTION
		on btn_openFile pressed do
		(
			--// Open a text file to add string items to use in the dropdown list
			if doesFileExist g_LBTools_quickSel.textFile then
			(
				ShellLaunch g_LBTools_quickSel.textFile ""
			)
			else
			(
				messagebox "Error: Text file does not exist. Recreating it." title:"Error"
				g_LBTools_quickSel.f_createPathsTextFile()
				g_LBTools_quickSel.f_loadPathItemsFromFile()
			)
		)
		
		on btn_refresh pressed do
		(
			--// Update the items list from the text file (assume they are each listed on a new line)
			if doesFileExist g_LBTools_quickSel.textFile then
			(
				g_LBTools_quickSel.f_loadPathItemsFromFile()
			)
			else
			(
				messagebox "Error: Text file does not exist. Recreating it." title:"Error"
				g_LBTools_quickSel.f_createPathsTextFile()
				g_LBTools_quickSel.f_loadPathItemsFromFile()
			)
		)
		
		on btn_grabPathObjs pressed do
		(
			g_LBTools_quickSel.f_selectPathObjects ddl_paths.selection
		)
		
		on ddl_paths selected val do
		(
			g_LBTools_quickSel.f_selectPathObjects val
		)		
	) --// end rollout
	
	rollout About_Rollout "About"
	(
		label lbl_author "Author - Lorne Brooks"
		hyperLink myEmail "lbtools2k@gmail.com" across:2 align:#center offset:[15,10] color:(color 0 0 255) hovercolor:(color 0 0 190) visitedcolor:(color 190 0 0) align:#center address:"mailto:lbtools2k@gmail.com"
		button btn_guide "" align:#right offset:[0,0] images:#(g_LBTools_quickSel.imageGuide, undefined, 1, 1, 1, 1, 1, false) tooltip:"Opens an online guide explaining the tool's features."
		label lbl_version "Version 1.01 - 12 Aug 2023"
		
		on btn_guide pressed do
		(
			shellLaunch "https://sites.google.com/view/lbtools-guide/home/quick-selection" ""
		)
	) --// end rollout
	

	-----------------------------------------------------------------------------------------
	--// Setup Interface
	-----------------------------------------------------------------------------------------
	--// initialize the script INI file for persistent data
	g_LBTools_quickSel.f_initializeIniFile()
	
	--// Create the paths text file if it doesn't exist
	g_LBTools_quickSel.f_createPathsTextFile()
	
	--// Get stored paths for dropdown list from text file
	g_LBTools_quickSel.f_loadPathItemsFromFile()
	
	--// Get stored selection sets for dropdown list from INI file
	g_LBTools_quickSel.f_loadSelectionSets()
	
	fn f_openQuickSelectionUI =
	(
		try(destroyDialog quickSelection_Rollout)catch()
		createDialog quickSelection_Rollout width:185 pos:g_LBTools_quickSel.QS_pos
	)
	
	f_openQuickSelectionUI()
)

) --// end struct

g_LBTools_quickSel = s_LBTools_quickSelection()


/*********************************************************************************************************************************************************************/
--// MACRO CALLS FOR UI
/*********************************************************************************************************************************************************************/

macroscript LBTools_quickSelection
	category:"LB Tools"
	icon:#("LBTools_QuickSelection",1)
	buttontext:"{ QS }"
	tooltip:"Quick Selection 1.01"

(
	g_LBTools_quickSel.f_quickSelection()
)

/*********************************************************************************************************************************************************************/